import logging

from modelforge.logs import setup_logging
from ast2vec import ensure_bblfsh_is_running_noexc, install_enry


__initialized__ = False


def initialize(log_level=logging.INFO, enry="./enry"):
    """
    Sets up the working environment: enables logging, launches the Babelfish
    server if it is not running, installs src-d/enry if it is not found in
    PATH.
    :param log_level: The verbosity level. Can be either an integer (e.g. \
    logging.INFO) or a string (e.g. "INFO").
    :param enry: The path to the linguist/enry executable. It if it exists,
    nothing happens. If it is not, src-d/enry is compiled into that file.
    :return:
    """
    global __initialized__
    if __initialized__:
        return
    setup_logging(log_level)
    ensure_bblfsh_is_running_noexc()
    install_enry(target=enry, warn_exists=False)
    __initialized__ = True



from setuptools import setup, find_packages


setup(
    name="tmsc",
    description="Part of source{d}'s stack for machine learning on source "
                "code. Provides API and tools to detect topics in Git "
                "repositories.",
    version="0.1.5-alpha",
    license="Apache Software License",
    author="source{d}",
    author_email="machine-learning@sourced.tech",
    url="https://github.com/src-d/tmsc",
    download_url="https://github.com/src-d/tmsc",
    packages=find_packages(exclude=("tmsc.tests",)),
    entry_points={
        "console_scripts": ["tmsc=tmsc.__main__:main"],
    },
    keywords=["machine learning on source code", "topic modeling",
              "github", "bblfsh", "babelfish", "ast2vec"],
    install_requires=["ast2vec>=0.3.8-alpha"],
    package_data={"": ["LICENSE.md", "README.md"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries"
    ]
)



import logging
import re

from ast2vec import Topics, Repo2Base, DocumentFrequencies
from ast2vec.bow import BOWBase
from ast2vec.model2.uast2bow import Uasts2BOW
from modelforge.backends import create_backend
import numpy
from scipy.sparse import csr_matrix

from tmsc.environment import initialize


class Repo2BOW(Repo2Base):
    """
    Implements the step repository -> :class:`ast2vec.nbow.NBOW`.
    """
    MODEL_CLASS = BOWBase

    def __init__(self, vocabulary, docfreq, **kwargs):
        super().__init__(**kwargs)
        self.vocabulary = vocabulary
        self._uasts2bow = Uasts2BOW(vocabulary, docfreq, lambda x: x.response.uast)

    def convert_uasts(self, file_uast_generator):
        return self._uasts2bow(file_uast_generator)


class TopicDetector:
    GITHUB_URL_RE = re.compile(
        r"(https://|ssh://git@|git://)(github.com/[^/]+/[^/]+)(|.git|/)")

    def __init__(self, topics=None, docfreq=None, bow=None, verbosity=logging.DEBUG,
                 prune_df_threshold=1, gcs_bucket=None, initialize_environment=True,
                 repo2bow_kwargs=None):
        if initialize_environment:
            initialize()
        self._log = logging.getLogger("topic_detector")
        self._log.setLevel(verbosity)
        if gcs_bucket:
            backend = create_backend(args="bucket=" + gcs_bucket)
        else:
            backend = create_backend()
        if topics is None:
            self._topics = Topics(log_level=verbosity).load(backend=backend)
        else:
            assert isinstance(topics, Topics)
            self._topics = topics
        self._log.info("Loaded topics model: %s", self._topics)
        if docfreq is None:
            if docfreq is not False:
                self._docfreq = DocumentFrequencies(log_level=verbosity).load(
                    source=self._topics.dep("docfreq")["uuid"], backend=backend)
            else:
                self._docfreq = None
                self._log.warning("Disabled document frequencies - you will "
                                  "not be able to query custom repositories.")
        else:
            assert isinstance(docfreq, DocumentFrequencies)
            self._docfreq = docfreq
        if self._docfreq is not None:
            self._docfreq = self._docfreq.prune(prune_df_threshold)
        self._log.info("Loaded docfreq model: %s", self._docfreq)
        if bow is not None:
            assert isinstance(bow, BOWBase)
            self._bow = bow
            if self._topics.matrix.shape[1] != self._bow.matrix.shape[1]:
                raise ValueError("Models do not match: topics has %s tokens while bow has %s" %
                                 (self._topics.matrix.shape[1], self._bow.matrix.shape[1]))
            self._log.info("Attached BOW model: %s", self._bow)
        else:
            self._bow = None
            self._log.warning("No BOW cache was loaded.")
        if self._docfreq is not None:
            self._repo2bow = Repo2BOW(
                {t: i for i, t in enumerate(self._topics.tokens)}, self._docfreq,
                **(repo2bow_kwargs or {}))
        else:
            self._repo2bow = None

    def query(self, url_or_path_or_name, size=5):
        if size > len(self._topics):
            raise ValueError("size may not be greater than the number of topics - %d" %
                             len(self._topics))
        if self._bow is not None:
            try:
                repo_index = self._bow.repository_index_by_name(
                    url_or_path_or_name)
            except KeyError:
                repo_index = -1
            if repo_index == -1:
                match = self.GITHUB_URL_RE.match(url_or_path_or_name)
                if match is not None:
                    name = match.group(2)
                    try:
                        repo_index = self._bow.repository_index_by_name(name)
                    except KeyError:
                        pass
        else:
            repo_index = -1
        if repo_index >= 0:
            token_vector = self._bow.matrix[repo_index]
        else:
            if self._docfreq is None:
                raise ValueError("You need to specify document frequencies model to process "
                                 "custom repositories")
            bow_dict = self._repo2bow.convert_repository(url_or_path_or_name)
            token_vector = numpy.zeros(self._topics.matrix.shape[1], dtype=numpy.float32)
            for i, v in bow_dict.items():
                token_vector[i] = v
            token_vector = csr_matrix(token_vector)
        topic_vector = -numpy.squeeze(self._topics.matrix.dot(token_vector.T).toarray())
        order = numpy.argsort(topic_vector)
        result = []
        i = 0
        while len(result) < size and i < len(self._topics):
            topic = self._topics.topics[order[i]]
            if topic:
                result.append((topic, -topic_vector[order[i]]))
            i += 1
        return result



import argparse
import json
import logging
import sys

from ast2vec import Topics, DocumentFrequencies, DEFAULT_BBLFSH_TIMEOUT
from ast2vec.bow import BOWBase
from modelforge.backends import create_backend

from tmsc.environment import initialize
from tmsc.topic_detector import TopicDetector


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Repository URL or path or name.")
    parser.add_argument("--log-level", default="INFO",
                        choices=logging._nameToLevel,
                        help="Logging verbosity.")
    parser.add_argument("--topics", default=None, help="Topic model URL or path.")
    parser.add_argument("--df", default=None,
                        help="Document frequencies URL or path.")
    parser.add_argument("--bow", default=None, help="BOW model URL or path.")
    parser.add_argument("--bblfsh", default=None,
                        help="babelfish server address.")
    parser.add_argument(
        "--timeout", type=int, default=None,
        help="Babelfish timeout - longer requests are dropped. Default is %s." %
             DEFAULT_BBLFSH_TIMEOUT)
    parser.add_argument("--gcs", default=None, help="GCS bucket to use.")
    parser.add_argument("--linguist", default=None,
                        help="Path to src-d/enry or github/linguist.")
    parser.add_argument("--prune-df", default=20, type=int,
                        help="Minimum number of times an identifer must occur in different "
                             "documents to be taken into account.")
    parser.add_argument("-n", "--nnn", default=10, type=int,
                        help="Number of topics to print.")
    parser.add_argument("-f", "--format", default="human", choices=["json", "human"],
                        help="Output format.")

    args = parser.parse_args()
    if args.linguist is None:
        args.linguist = "./enry"
    initialize(args.log_level, enry=args.linguist)
    if args.gcs:
        backend = create_backend(args="bucket=" + args.gcs)
    else:
        backend = create_backend()
    if args.topics is not None:
        args.topics = Topics(log_level=args.log_level).load(source=args.topics, backend=backend)
    if args.df is not None:
        args.df = DocumentFrequencies(log_level=args.log_level).load(
            source=args.df, backend=backend)
    if args.bow is not None:
        args.bow = BOWBase(log_level=args.log_level).load(source=args.bow, backend=backend)
    sr = TopicDetector(
        topics=args.topics, docfreq=args.df, bow=args.bow, verbosity=args.log_level,
        prune_df_threshold=args.prune_df, gcs_bucket=args.gcs, repo2bow_kwargs={
            "linguist": args.linguist,
            "bblfsh_endpoint": args.bblfsh,
            "timeout": args.timeout})
    topics = sr.query(args.input, size=args.nnn)
    if args.format == "json":
        json.dump({"repository": args.input, "topics": topics}, sys.stdout)
    elif args.format == "human":
        for t, r in topics:
            print("%64s" % t, "%.2f" % r, sep="\t")


if __name__ == "__main__":
    sys.exit(main())
